"""Common GRPC backend client."""
import logging
import random
import string
import time
import warnings
from types import ModuleType
from typing import Any, Callable, Dict, Generator, Iterable, Sequence

import grpc
from google.protobuf.empty_pb2 import Empty
from google.protobuf.message import Message
from google.protobuf.service import Service
from google.protobuf.text_format import MessageToString

from .decoder import GrpcDecoder
from .exceptions import ServiceNotFound

_LOGGER = logging.getLogger(__name__)


class CallWrapper:
    """
    Wrapper for gRPC calls.

    Arguments:
        call: A gRPC method to be called.
        request: A gRPC request to be sent.
        client: A GrpcClient instance.
        timeout: Timeout for the call.
        max_retries: A maximum number of calls to be made.
        modify_result: A callable to modify the result of the gRPC method.
    """

    def __init__(self, call: Callable, request: Message, *, client: 'GrpcClient', timeout: int = None,
                 max_retries: int = 10, modify_result: Callable[[Any], Any] = None):
        self.call = call
        self.request = request
        self.client = client
        self.timeout = timeout
        self.max_retries = max_retries
        self.modify_result = modify_result
        self._call_id = ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(4))

    def get_reply(self) -> Any:
        """Fetch and return the reply."""
        return self.call(self.request, timeout=self.timeout)

    def decode_reply(self, reply) -> Any:
        """Decode reply and return decoded result or raise decoded exception."""
        # Assert the supported structure of messages
        assert ('exception' in reply.DESCRIPTOR.fields_by_name or 'data' in reply.DESCRIPTOR.fields_by_name
                or isinstance(reply, Empty))

        modify_result = self.modify_result or (lambda x: x)

        # Decode exception first
        if 'exception' in reply.DESCRIPTOR.fields_by_name and reply.HasField('exception'):
            raise self.client.decoder.decode_exception(reply.exception)  # type: ignore

        # Decode data otherwise
        if 'data' in reply.DESCRIPTOR.fields_by_name and reply.HasField('data'):
            return modify_result(self.client.decoder.decode(reply.data))  # type: ignore

        # The reply is empty
        return modify_result(None)

    def reconnect(self) -> None:
        """Try to reconnect the client if unavailable."""
        self.client.connect()

    def result(self) -> Any:
        """Return the call result."""
        # XXX: Extract the name of the gRPC method
        method = self.call._method.decode('utf-8')  # type: ignore
        _LOGGER.debug('[%s] %s(%s)', self._call_id, method, MessageToString(self.request, as_one_line=True))

        retries = 0
        try:
            while True:
                time.sleep(0.1 * retries)  # Wait for increasing amount of time before each call
                try:
                    retries += 1
                    reply = self.get_reply()
                    break
                except grpc.RpcError as error:
                    if error.code() == grpc.StatusCode.UNAVAILABLE and retries < self.max_retries:
                        # Connection probably failed. Try to reconnect.
                        _LOGGER.info('[%s] Connection failed. Trying to reconnect.', self._call_id)
                        self.reconnect()
                    else:
                        raise
        except grpc.RpcError as error:
            _LOGGER.info('[%s] %s failed with %s', self._call_id, method, error)
            raise

        # Streaming returns a sequence of responses which can't be logged.
        if isinstance(reply, Message):
            result_repr = MessageToString(reply, as_one_line=True)
        else:
            result_repr = 'a response'
        _LOGGER.debug('[%s] %s returned %s', self._call_id, method, result_repr)

        return self.decode_reply(reply)


class StreamCallWrapper(CallWrapper):
    """Wrapper for gRPC call with streaming response."""

    def decode_reply(self, reply: Iterable[Any]) -> Generator[Any, None, None]:
        """Decode replies one by one."""
        for item in reply:
            yield super().decode_reply(item)


class FutureCallWrapper(CallWrapper):
    """Wrapper for gRPC future call."""

    def __init__(self, call: Callable, request: Message, *, client: 'GrpcClient', timeout: int = None,
                 max_retries: int = 10, modify_result: Callable[[Any], Any] = None):
        super().__init__(call, request, client=client, timeout=timeout, max_retries=max_retries,
                         modify_result=modify_result)
        # Get a future object - start an asynchronous request
        self._future = self._call_future()

    def _call_future(self) -> grpc.Future:
        """Start asynchronous request and return a future object."""
        future_call = self.call.future  # type: ignore
        return future_call(self.request, timeout=self.timeout)

    def get_reply(self) -> Any:
        """Fetch and return the reply."""
        return self._future.result()

    def reconnect(self) -> None:
        """Try to reconnect the client if unavailable."""
        # Cancel old request
        self._future.cancel()
        # Reconnect
        super().reconnect()
        # Start a new asynchronous request.
        self._future = self._call_future()

    def __getattr__(self, name):
        """Proxy all other future properties."""
        return getattr(self._future, name)


class GrpcClient(object):
    """GRPC backend client."""

    def __init__(self, netloc: str, decoder: GrpcDecoder, grpc_modules: Sequence[ModuleType],
                 credentials: grpc.ChannelCredentials = None):
        """
        Initialize GRPC backend client instance.

        Args:
            netloc: Network location of a gRPC server.
            decoder: Decoder instance.
            grpc_modules: List of generated _pb2_grpc modules with service stubs.
            credentials: Credentials for a secure channel connection. If None, insecure channel is used.
        """
        self.netloc = netloc
        self.credentials = credentials
        self.decoder = decoder
        self.grpc_modules = grpc_modules

        # Stubs cache
        self._stubs = {}  # type: Dict[str, Any]

        self.connect()

    def connect(self) -> None:
        """Create channel to GPRC server."""
        if self.credentials is not None:
            self.channel = grpc.secure_channel(self.netloc, self.credentials)
        else:
            self.channel = grpc.insecure_channel(self.netloc)

    def get_stub(self, service: str) -> Service:
        """
        Return stub for the service.

        Stubs are cached, so for each service, there is only one stub created.
        """
        stub_name = '{}Stub'.format(service)

        if stub_name in self._stubs:
            return self._stubs[stub_name]

        for module in self.grpc_modules:
            if hasattr(module, stub_name):
                stub_cls = getattr(module, stub_name)
                stub = stub_cls(self.channel)
                self._stubs[stub_name] = stub
                return stub
        else:
            raise ServiceNotFound('Service {} could not be found'.format(stub_name))

    def call(
        self, service: str, method: str, request: Message, *,
        blocking: bool = None, max_retries: int = 10, modify_result: Callable[[Any], Any] = None
    ) -> Any:
        """
        Call grpc method.

        Args:
            service: Name of service, that is used to get service stub.
            method: Name of service method.
            request: Request message passed as an argument to service method.
            blocking: Determines whether to use blocking or non-blocking call.
            max_retries: A maximum number of calls to be made.
            modify_result: Callable that processes decoded result and modifies it. Default is identity.
        """
        if blocking is not None:
            if blocking:
                warnings.warn("Argument blocking is deprecated and will be deleted.", DeprecationWarning)
            else:
                warnings.warn("Argument blocking is deprecated, use method call_future instread.", DeprecationWarning)
                return self.call_future(service, method, request, max_retries=max_retries, modify_result=modify_result)
        stub = self.get_stub(service)
        grpc_method = getattr(stub, method)
        wrapper = CallWrapper(grpc_method, request, client=self, max_retries=max_retries, modify_result=modify_result)
        return wrapper.result()

    def call_future(self, service: str, method: str, request: Message, *, max_retries: int = 10,
                    modify_result: Callable[[Any], Any] = None) -> FutureCallWrapper:
        """
        Call grpc method asynchronously.

        Args:
            service: Name of service, that is used to get service stub.
            method: Name of service method.
            request: Request message passed as an argument to service method.
            max_retries: A maximum number of calls to be made.
            modify_result: Callable that processes decoded result and modifies it. Default is identity.
        """
        stub = self.get_stub(service)
        grpc_method = getattr(stub, method)
        return FutureCallWrapper(grpc_method, request, client=self, max_retries=max_retries,
                                 modify_result=modify_result)

    def call_stream(self, service: str, method: str, request: Message, *, max_retries: int = 10,
                    modify_result: Callable[[Any], Any] = None) -> Generator[Any, None, None]:
        """
        Call grpc method with streaming response.

        Args:
            service: Name of service, that is used to get service stub.
            method: Name of service method.
            request: Request message passed as an argument to service method.
            max_retries: A maximum number of calls to be made.
            modify_result: Callable that processes decoded result and modifies it. Default is identity.
        """
        stub = self.get_stub(service)
        grpc_method = getattr(stub, method)
        wrapper = StreamCallWrapper(grpc_method, request, client=self, max_retries=max_retries,
                                    modify_result=modify_result)
        return wrapper.result()
