from functools import partial
from typing import cast
from unittest import TestCase
from unittest.mock import Mock, call, sentinel

from frgal.client import GrpcClient
from frgal.decoder import GrpcDecoder
from frgal.utils import CallableMock, MockingChannel, TestClientMixin

from .proto.example_pb2 import Reply, SimpleMessage


class CallableMockTest(TestCase):
    def test_call(self):
        a_callable = Mock(return_value=sentinel.reply)
        a_call = CallableMock('/Coffee/make', a_callable)

        self.assertEqual(a_call(sentinel.request), sentinel.reply)

        self.assertEqual(a_callable.mock_calls, [call(sentinel.request)])


class MockingChannelTest(TestCase):
    def _test_callable(self, name: str):
        client = Mock(name='client')
        channel = MockingChannel(client)

        a_callable = getattr(channel, name)('/Coffee/make')

        a_call = cast(partial, a_callable.call)
        self.assertEqual(a_call.func, client.mock)
        self.assertEqual(a_call.args, ())
        self.assertEqual(a_call.keywords, {'method': '/Coffee/make'})

    def test_unary_unary(self):
        self._test_callable(name='unary_unary')

    def test_unary_stream(self):
        self._test_callable(name='unary_stream')

    def test_stream_unary(self):
        self._test_callable(name='stream_unary')

    def test_stream_stream(self):
        self._test_callable(name='stream_stream')


class CoffeeStub:
    """A simple stub for testing."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.channel = channel
        self.make = channel.unary_unary('/Coffee/make')


class TestClient(TestClientMixin, GrpcClient):
    """Simple test client to test a TestClientMixin."""


class TestClientMixinTest(TestCase):
    def test_connect(self):
        client = TestClient(sentinel.netloc, sentinel.decoder, [])

        client.connect()

        self.assertIsInstance(client.channel, MockingChannel)

    def test_call(self):
        client = TestClient(sentinel.netloc, GrpcDecoder(), [Mock(CoffeeStub=CoffeeStub)])
        request = SimpleMessage(value='10')
        reply = Reply()
        reply.data.result = 'Espresso!'
        client.mock.return_value = reply

        result = client.call('Coffee', 'make', request)

        self.assertEqual(result, 'Espresso!')
        self.assertEqual(client.mock.mock_calls, [call(request, method='/Coffee/make', timeout=None)])

    def test_call_future(self):
        client = TestClient(sentinel.netloc, GrpcDecoder(), [Mock(CoffeeStub=CoffeeStub)])
        request = SimpleMessage(value='10')
        reply = Reply()
        reply.data.result = 'Espresso!'
        client.mock.return_value = reply

        result = client.call_future('Coffee', 'make', request)

        self.assertEqual(result.result(), 'Espresso!')
        self.assertEqual(client.mock.mock_calls, [call(request, method='/Coffee/make', timeout=None)])

    def test_call_stream(self):
        client = TestClient(sentinel.netloc, GrpcDecoder(), [Mock(CoffeeStub=CoffeeStub)])
        request = SimpleMessage(value='10')
        reply = Reply()
        reply.data.result = 'Espresso!'
        client.mock.return_value = [reply, reply]

        result = client.call_stream('Coffee', 'make', request)

        self.assertEqual(tuple(result), ('Espresso!', 'Espresso!'))
        self.assertEqual(client.mock.mock_calls, [call(request, method='/Coffee/make', timeout=None)])
